/*
 * bdevid.c - unique id probe for ata devices
 * 
 * Copyright 2005 Peter M. Jones
 * Copyright 2005 Red Hat, Inc.
 * portions from udev's scsi_serial.c Copyright 2003 IBM Corp.
 * 
 * vim:ts=8:sw=8:sts=8:noet
 *
 */

#define _GNU_SOURCE 1

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <scsi/scsi.h>
#include <scsi/sg.h>

#define ARGHA _POSIX_C_SOURCE
#undef _POSIX_C_SOURCE
#define ARGHB _GNU_SOURCE
#undef _GNU_SOURCE
#include <Python.h>
#undef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE ARGHA
#undef ARGHA
#undef _GNU_SOURCE
#define _GNU_SOURCE ARGHB
#undef ARGHB

/* this should wind up just like zlib's, but with sane types */
extern u_int32_t adler32(u_int32_t adler, const unsigned char *buf, int len);

static const char hex_str[]="0123456789abcdef";

#define SCSI_INQ_BUFF_LEN	254

#define SCSI_ID_VENDOR_SPECIFIC	0
#define SCSI_ID_T10_VENDOR	1
#define SCSI_ID_EUI_64		2
#define SCSI_ID_NAA		3

#define	SCSI_ID_NAA_DONT_CARE		0xff
#define	SCSI_ID_NAA_IEEE_REG		5
#define	SCSI_ID_NAA_IEEE_REG_EXTENDED	6

#define	SCSI_ID_BINARY	1
#define	SCSI_ID_ASCII	2

struct scsi_id_search_values {
	u_char	id_type;
	u_char	naa_type;
	u_char	code_set;
};

static const struct scsi_id_search_values id_search_list[] = {
	{ SCSI_ID_NAA, SCSI_ID_NAA_IEEE_REG_EXTENDED, SCSI_ID_BINARY },
	{ SCSI_ID_NAA, SCSI_ID_NAA_IEEE_REG_EXTENDED, SCSI_ID_ASCII },
	{ SCSI_ID_NAA, SCSI_ID_NAA_IEEE_REG, SCSI_ID_BINARY },
	{ SCSI_ID_NAA, SCSI_ID_NAA_IEEE_REG, SCSI_ID_ASCII },
	{ SCSI_ID_NAA, SCSI_ID_NAA_DONT_CARE, SCSI_ID_BINARY },
	{ SCSI_ID_NAA, SCSI_ID_NAA_DONT_CARE, SCSI_ID_ASCII },
	{ SCSI_ID_EUI_64, SCSI_ID_NAA_DONT_CARE, SCSI_ID_BINARY },
	{ SCSI_ID_EUI_64, SCSI_ID_NAA_DONT_CARE, SCSI_ID_ASCII },
	{ SCSI_ID_T10_VENDOR, SCSI_ID_NAA_DONT_CARE, SCSI_ID_BINARY },
	{ SCSI_ID_T10_VENDOR, SCSI_ID_NAA_DONT_CARE, SCSI_ID_ASCII },
	{ SCSI_ID_VENDOR_SPECIFIC, SCSI_ID_NAA_DONT_CARE, SCSI_ID_BINARY },
	{ SCSI_ID_VENDOR_SPECIFIC, SCSI_ID_NAA_DONT_CARE, SCSI_ID_ASCII },
};
static const int id_search_list_nents =
	sizeof(id_search_list)/sizeof(id_search_list[0]);



#define DID_OK		0x00	/* NO error                                */
#define DID_NO_CONNECT	0x01	/* Couldn't connect before timeout period  */
#define DID_BUS_BUSY	0x02	/* BUS stayed busy through time out period */
#define DID_TIME_OUT	0x03	/* TIMED OUT for other reason              */

#define DRIVER_TIMEOUT	0x06
#define DRIVER_SENSE	0x08       /* Sense_buffer has been set */

#define VENDOR_LENGTH	8
#define MODEL_LENGTH	16

#define SCSI_CHECK_CONDITION	0x2
#define SCSI_COMMAND_TERMINATED	0x22

#define SENSE_BUFF_LEN		32
#define SCSI_INQ_BUFF_LEN	254
#define INQUIRY_CMD	0x12
#define INQUIRY_CMDLEN	6

#define DEF_TIMEOUT	5000

static int
scsi_check_error(struct sg_io_hdr *sg)
{
	int scsi_status = sg->status & 0x7e;
	int msg_status = sg->msg_status;
	int host_status = sg->host_status;
	int driver_status = sg->driver_status;
	unsigned char *sb = sg->sbp;
	int sb_len = sg->sb_len_wr;
	
	if (!scsi_status && !msg_status && !host_status && !driver_status)
		return 0;

	if (scsi_status == SCSI_CHECK_CONDITION ||
			scsi_status == SCSI_COMMAND_TERMINATED ||
			(driver_status & 0xf) == DRIVER_SENSE) {
		if (sb && sb_len > 2) {
			int sense_key;
			unsigned char asc;

			if (sb[0] & 0x2) {
				sense_key = sb[1] & 0xf;
				asc = sb[2];
			} else {
				sense_key = sb[2] & 0xf;
				asc = (sb_len > 12) ? sb[12] : 0;
			}

			switch (sense_key) {
				case RECOVERED_ERROR:
					return 0;
				case UNIT_ATTENTION:
					if (asc == 0x28 || asc == 0x29)
						return -EAGAIN;
					return -EINVAL;
				case ILLEGAL_REQUEST:
					return -EINVAL;
			}
		}
		return -EINVAL;
	}
	if (host_status == DID_NO_CONNECT || host_status == DID_BUS_BUSY ||
			host_status == DID_TIME_OUT)
		return -EAGAIN;
	if (driver_status == DRIVER_TIMEOUT)
		return -EAGAIN;
	return -EIO;
}

static int
scsi_inquiry(const int fd, unsigned char *buf, u_int8_t buflen,
		u_int8_t evpd, u_int8_t page)
{
	u_int8_t len = buflen > SCSI_INQ_BUFF_LEN ? SCSI_INQ_BUFF_LEN : buflen;
	unsigned char cmd[INQUIRY_CMDLEN] = {
		INQUIRY_CMD, evpd, page, 0, len, 0 };
	unsigned char sense[SENSE_BUFF_LEN];
	struct sg_io_hdr sg = {
		.interface_id = 'S',
		.cmd_len = sizeof (cmd),
		.mx_sb_len = sizeof (sense),
		.dxfer_direction = SG_DXFER_FROM_DEV,
		.dxfer_len = len,
		.dxferp = buf,
		.cmdp = cmd,
		.sbp = sense,
		.timeout = DEF_TIMEOUT,
	};
	int tries = 3;

	while (tries--) {
		int rc;

		rc = ioctl(fd, SG_IO, &sg);
		if (rc < 0) {
			if (errno == EAGAIN)
				continue;
			return -errno;
		}
			
		rc = scsi_check_error(&sg);
		if (rc == -EAGAIN)
			continue;
		return rc;
	}
	return -EIO;
}

static int
get_vendor_model(const int fd, char *buf, const int len)
{

	unsigned char page[SCSI_INQ_BUFF_LEN] = { 0 };
	int rc;
	
	rc = scsi_inquiry(fd, page, sizeof (page), 1, 0x83);
	if (rc < 0)
		return rc;

	memcpy(buf, page+8, VENDOR_LENGTH);
	buf[VENDOR_LENGTH]='\0';
	buf = buf + strlen(buf);
	memcpy(buf, page+16, MODEL_LENGTH);
	buf[VENDOR_LENGTH+MODEL_LENGTH]='\0';

	return 0;
}

static int
scsi_page83_inquiry(const int fd, char **serial)
{
	unsigned char page[SCSI_INQ_BUFF_LEN] = { 0 };
	int rc, i;
	
	rc = scsi_inquiry(fd, page, sizeof (page), 1, 0x83);
	if (rc < 0)
		return rc;

	if (page[1] != 0x83)
		return -EIO;

	for (i = 0; i < id_search_list_nents ; i++) {
		struct scsi_id_search_values *id =
			(struct scsi_id_search_values *)&id_search_list[i];
		char *buf;
		int j, len;
			
		for (j = 4; j <= page[3] + 3; j += page[j + 3] + 4) {
			u_int32_t adler;

			if (page[j+1] & 0x30)
				continue;
			if ((page[j+1] & 0x0f) != id->id_type)
				continue;
			if (id->naa_type != SCSI_ID_NAA_DONT_CARE &&
					id->naa_type != ((page[j+4]&0xf0) >> 4))
				continue;

			if ((page[j] & 0x0f) != id->code_set)
				continue;

			len = page[j+3];
			if ((page[j] & 0x0f) != SCSI_ID_ASCII)
				len *= 2;

			len += 2;
			if (id->id_type == SCSI_ID_VENDOR_SPECIFIC)
				len += VENDOR_LENGTH + MODEL_LENGTH;

			buf = calloc(len, sizeof (char));
			if (buf == NULL)
				return -errno;

			buf[0] = hex_str[id->id_type];

			get_vendor_model(fd, buf+1, len-2);

			if (page[j] == SCSI_ID_ASCII) {
				char *s = buf + strlen(buf);
				unsigned char *idtext = &page[j+4];

				strncpy(s, (char *)idtext, 4 + page[j+3]);
				if (!strcmp(s, "Linux ATA-SCSI simulator")) {
					free(buf);
					return 0;
				}
			} else {
				char *s = buf + strlen(buf);
				unsigned char *idbin = &page[j+4];
				int k;

				for (k = 0; k < 4 + page[j+3]; k++) {
					*(s++) = hex_str[(*idbin & 0xf0) >> 4];
					*s = hex_str[*(idbin++) & 0x0f];
				}
			}

			adler = adler32((u_int32_t)0, NULL, 0);
			adler = adler32(adler, (unsigned char *)buf,
					strlen(buf));

			sprintf(buf, "%08" PRIx32, adler);
			*serial = buf;
			return 1;
		}
	}
	return 0;
}

static int
scsi_page80_inquiry(const int fd, char **serial)
{
	unsigned char page[SCSI_INQ_BUFF_LEN] = { 0 };
	int rc, len;
	u_int32_t adler;
	char *buf;
	
	rc = scsi_inquiry(fd, page, sizeof (page), 1, 0x80);
	if (rc < 0)
		return rc;

	if (page[1] != 0x80)
		return -EIO;

	len = 1 + VENDOR_LENGTH + MODEL_LENGTH + page[3];

	buf = calloc(len, sizeof (char));
	if (buf < 0)
		return -errno;

	buf[0] = 'S';

	get_vendor_model(fd, buf+1, len-2);

	strncpy(buf + strlen(buf), (char *)page+4, page[3] + 4);

	adler = adler32((u_int32_t)0, NULL, 0);
	adler = adler32(adler, (unsigned char *)buf, strlen(buf));

	sprintf(buf, "%08" PRIx32, adler);
	*serial = buf;
	return 1;
}

static int
scsi_check(const int fd)
{
	unsigned char page[SCSI_INQ_BUFF_LEN] = { 0 };
	int rc;

	rc = scsi_inquiry(fd, page, sizeof(page), 1, 0x0);
	if (rc < 0)
		return 0;
	return 1;
}

static int
scsi_get_device_serial(const int fd, char **serial)
{
	unsigned char page[SCSI_INQ_BUFF_LEN] = { 0 };
	int i, rc;
	int has80=0, has83=0;

	rc = scsi_inquiry(fd, page, sizeof (page), 1, 0x0);
	if (rc < 0)
		return 0;

	for (i = 4; i <= page[3] + 3; i++) {
		if (page[i] == 0x83)
			has83 = 1;
		if (page[i] == 0x80)
			has80 = 1;
	}

	memset(page, 0, sizeof (page));
	if (has83) {
		rc = scsi_page83_inquiry(fd, serial);
		if (rc <= 0 && has80)
			rc = scsi_page80_inquiry(fd, serial);
	} else if (has80)
		rc = scsi_page80_inquiry(fd, serial);
	return rc;
}

static PyObject *
pyb_get_serial(PyObject *self, PyObject *args, PyObject *kwds)
{
	char *kwlist[] = {"path", NULL};
	char *path = NULL;
	char *serial = NULL;
	PyObject *obj = NULL;
	int fd;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s:scsi_id",
			kwlist, &path))
		return NULL;

	fd = open(path, O_RDONLY);
	if (fd < 0 || !scsi_check(fd) || !scsi_get_device_serial(fd, &serial)) {
		PyErr_SetString(PyExc_ValueError, "not a scsi device");
		goto out;
	}

	obj = PyString_FromString(serial);
	free(serial);
out:
	if (fd > 0)
		close(fd);
	return obj;
}

static PyMethodDef pyb_functions[] = {
	{"scsi_id", (PyCFunction)pyb_get_serial, METH_VARARGS|METH_KEYWORDS},
	{NULL},
};

PyMODINIT_FUNC
initbdevid(void)
{
	Py_InitModule("bdevid", pyb_functions);
}
